local core=require("core")
function core.credits(...)
  for k,v in ipairs(core.authors) do
    core.message(v,...)
  end
end

function core.scrsh(n)
  local scr=love.graphics.newScreenshot()
  if n then
    n2=n/255
    scr:mapPixel(function(x,y,r,b,g,a) return r*n2,b*n2,g*n2,a end)
  end
  return love.graphics.newImage(scr)
end
function core.fadeout(color)
  local color=color or {0,0,0,255}
  core.reset()
  local x=0
  local bg=core.scrsh()
  function love.draw()
    love.graphics.setColor(255,255,255,255)
    love.graphics.draw(bg,0,0,0,love.graphics.getWidth()/bg:getWidth(),love.graphics.getHeight()/bg:getHeight())
  end
  core.step()
  function love.draw()
    local r,g,b,a=unpack(color)
    love.graphics.setColor(255,255,255,255)
    love.graphics.draw(bg,0,0,0,love.graphics.getWidth()/bg:getWidth(),love.graphics.getHeight()/bg:getHeight())
    love.graphics.setColor(r,g,b,a*x)
    love.graphics.setRectangleTexture()
    love.graphics.rectangle("fill",0,0,love.graphics.getWidth(),love.graphics.getHeight())
  end
  function love.update(dt)
    x=x+dt
  end
  while true do
    core.step()
    if x >= 1 then
      love.graphics.setColor(255,255,255,255)
      core.reset()
      return
    end
  end
end
function core.randomid()
  local res=""
  for n=1,10 do
    local x = math.random(1,#core.idcharset)
    res=res..core.idcharset:sub(x,x)
  end
  return res
end
core.touchmegaid=core.randomid()
function core.deleteline(line)
  local fl1,fl2=line:find("\n")
  if fl1 and fl2 then
  return line:sub(fl2+1)
  else
  return line
  end
end
function core.linecount(line)
  local res=0
  for x in string.gmatch(line,"\n") do
    res=res+1
  end
  return res
end
function love.errhand(err)
  pcall(love.window.showMessageBox,"Crash!",debug.traceback(err,2),"error")
end
function core.wrap(wfunc,mfunc)
  local func=wfunc
  return function(...)
    return mfunc(...)
  end
end
function core.isInside(x,y,rectx,recty,rectw,recth)
  if x > rectx and y > recty and x < rectx+rectw and y < recty+recth then
    return true
  end
  return false
end
function core.resize(img,sizex,sizey)
  local sizex=sizex or 1
  local sizey=sizey or sizex
  local canvas=love.graphics.newCanvas(img:getWidth()*sizex,img:getHeight()*(sizey))
  canvas:renderTo(function()
    core.origin()
    love.graphics.draw(img,0,0,0,sizex,sizey)
  end)
  return love.graphics.newImage(canvas:newImageData())
end
function core.step()
  coroutine.yield()
end
function core.dat(type,msg)
  return "["..os.date().."]["..(type or "MAIN THREAD/INFO").."] "..(msg or "")
end
function core.setFont(font,size)
  core.font=font
  core.fontSize=size
end
function core.message(msg,bg,color,t)
  core.reset()
  local End=false
  -- enable key repeat so backspace can be held down to trigger love.keypressed multiple times.
  love.keyboard.setKeyRepeat(true)
  function love.keypressed(key)
    if key=="return" and not t then
      End=true
    end
  end
  function love.mousepressed()
    if not t then
      End=true
    end
  end
  function love.update(dt)
    if t then
      t=t-dt
    end
    if t and t <= 0 then
      End=true
    end
  end
  function love.draw()
    local height = 0
    local lines = 0
    local texth=0
    if core.font then
      height = core.font:getHeight( )
      local width,l=core.font:getWrap((msg or tostring(msg) or "NULL MESSAGE"),love.graphics.getWidth()/2*core.getFontScale(core.font))
      lines=#l
      texth=lines*24
    end
    if bg then
      if color then
        love.graphics.setColor(color,color,color)
      end
      love.graphics.draw(bg,0,0,0,love.graphics.getWidth()/bg:getWidth(),love.graphics.getHeight()/bg:getHeight())
    end
    if core.font then
      love.graphics.setFont(core.font)
    end
    love.graphics.setColor(255,255,255,255)
    love.graphics.printf((msg or tostring(msg) or "NULL MESSAGE"), 0, (love.graphics.getHeight()-texth)/2, love.graphics.getWidth()/2,"center",0,2)
  end
  while true do
    core.step()
    if End==true then
      core.reset()
      return
    end
  end
end
function core.read(msg,def,bg,color)
  local utf8 = require("utf8")
  local text = def or ""
  local End=false
  local toblink=0
  local curvisible=true
  -- enable key repeat so backspace can be held down to trigger love.keypressed multiple times.
  love.keyboard.setTextInput(true)
  love.keyboard.setKeyRepeat(true)
  
  function love.textinput(t)
    text = text .. t
  end
  function love.update(dt)
    toblink=toblink-dt
    if toblink <= 0 then
      toblink=toblink+0.5
      curvisible=not curvisible
    end
  end
  function love.keypressed(key)
    if key == "backspace" then
      -- get the byte offset to the last UTF-8 character in the string.
      local byteoffset = utf8.offset(text, -1)
  
      if byteoffset then
        -- remove the last UTF-8 character.
        -- string.sub operates on bytes rather than UTF-8 characters, so we couldn't do string.sub(text, 1, -2).
        text = string.sub(text, 1, byteoffset - 1)
      end
    end
    if key=="return" then
      End=true
    end
  end
  function love.mousereleased()
    love.keyboard.setTextInput(not love.keyboard.hasTextInput())
  end
  function love.draw()
    local height = 0
    local lines = 0
    local texth=0
    if core.font then
      height = core.font:getHeight( )
      local width,l=core.font:getWrap((msg or tostring(msg) or "Enter a text: ")..text..((curvisible and love.keyboard.hasTextInput()) and "|" or ""),love.graphics.getWidth()/2*core.getFontScale(core.font))
      lines=#l
      texth=lines*24
    end
    if bg then
      if color then
        love.graphics.setColor(color,color,color)
      end
      love.graphics.draw(bg,0,0,0,love.graphics.getWidth()/bg:getWidth(),love.graphics.getHeight()/bg:getHeight())
    end
    if core.font then
      love.graphics.setFont(core.font)
    end
    love.graphics.setColor(255,255,255,255)
    --print(texth)
    love.graphics.printf((msg or tostring(msg) or "Enter a text: ")..text..((curvisible and love.keyboard.hasTextInput()) and "|" or ""), 0, 10, love.graphics.getWidth()/2,"center",0,2)
  end
  while true do
    core.step()
    if End then
      core.reset()
      love.keyboard.setTextInput(false)
      return text
    end
  end
end
core.handlers={}
core.hasCursor = love.mouse.hasCursor
function love.mouse.hasCursor()
  if arg[1]=="-gamepad" or core.gamepad then
    return false
  else
    return core.hasCursor()
  end
end
function core.handlers.threaderror(thread,err)
  if core.error then
    core.error:play()
  end
  print("[THREAD UNHANDLED ERROR] "..err)
  love.window.showMessageBox("Error",err,"error")
end
function core.handlers.keypressed(key)
--  if core.click then
--    core.click:rewind()
--    core.click:play()
--  end
  if key=="f11" then
    love.window.setFullscreen(not love.window.getFullscreen())
  end
--  print("USER PRESSED KEY_"..key)
end
function core.reset()
  for n,s in pairs(love.handlers) do
    love[n] = nil
    love.draw = nil
    love.update = nil
    love.load = nil
  end
end
function core.newSource(fname,channels,size)
  local size = size or 1
  local snddata1=love.sound.newSoundData(fname)
  local snddata2=love.sound.newSoundData(snddata1:getSampleCount()*size,snddata1:getSampleRate(),snddata1:getBitDepth(),channels)
  local supersize=snddata1:getSampleCount()
  for x=0,size-1 do
    for n=0,math.min(snddata2:getSampleCount(),snddata1:getSampleCount())-1 do
      snddata2:setSample(n+(x*supersize),snddata1:getSample(n))
    end
  end
  local res= love.audio.newSource(snddata2)
  return res
end
function core.custommenu(buttons,title,sel,add)
  core.reset()
  local add= add or {}
  local bg=add.bg
  local scale=add.scale or 0.55
  local func=add.func
  local rectbg=add.rectbg
  local bgcolor=add.bgcolor
  local btncolora=add.btncolora
  local btncoloria=add.btncoloria
  local btntitlecolor=add.btntitlecolor
  local titlecolor=add.titlecolor
  local btncolorina=add.btncolorina
  --local scalex,scaley=add.scalex or add.scale or 1,add.scaley or add.scalex
  local offset=add.offset or 0
  love.keyboard.setKeyRepeat(true)
  local End = false
  local active=sel or 1
  function love.keypressed(key)
    --if key=="up" then
    --  if active-1 < 1 then
    --    active = #buttons
    --  else
    --    active = active-1
    --  end
    --  while buttons[active].ina do
    --  if active-1 < 1 then
    --    active = #buttons
    --  else
    --    active = active-1
    --  end
    --  end
    --end
    --if key=="down" then
    --  if active+1 > #buttons then
    --    active = 1
    --  else
    --    active = active+1
    --  end
    --  while buttons[active].ina do
    --  if active+1 > #buttons then
    --    active = 1
    --  else
        -- active = active+1
      -- end
      -- end
    -- end
    -- if key=="return" then
        -- if core.click then
      -- core.click:stop()
      -- core.click:play()
    -- end
      -- End=true
    -- end
  end
  local function getkey(x,y,wx,hx,w,h)
    local areaX=w/wx
    local areaY=h/hx
    local function clamp(x,c1,c2)
      return math.min(c2,math.max(x,c1))
    end
    local rx,ry=math.floor(clamp(x,0,w-1)/areaX)+1,math.floor(clamp(y,0,h-1)/areaY)+1
    return rx,ry
  end
  local mouseOnButton=false
  local mouseClicked=false
  function love.mousepressed(x,y)
    if mouseOnButton and not buttons[active].ina then
      mouseClicked=active
    end
  end
  function love.mousereleased(x,y)
    if active and mouseClicked and mouseClicked==active and not buttons[active].ina then
      if core.click then
        core.click:stop()
        core.click:play()
      end
      End=true
    end
    mouseClicked=false
  end
  local nocolor={0,0,0,255}
  function love.update(dt)
    local x,y=love.mouse.getPosition()
    local btns={}
    local hb=false
    for n,s in ipairs(buttons) do
      if core.isInside(x/scale,y/scale,10,n*200-60+offset,love.graphics.getWidth()/scale-20,200) then
        active=n
        hb=true
      end
    end
    mouseOnButton=hb
    if active then
      if (active+2)*200+60+offset>love.graphics.getHeight()/scale then
        offset=offset-800*dt end
      if (active)*200+60+offset<200+60 then
        offset=offset+800*dt
      end
    end
  end
  function love.draw()
    --love.graphics.scale(scalex or 1,scaley or 1)
    --if buttons[active].ina then
    --  if active+1 > #buttons then
    --    active = 1
    --  else
    --    active = active+1
    --  end
    --  while buttons[active].ina do
    --    if active+1 > #buttons then
    --      active = 1
    --    else
    --      active = active+1
    --    end
    --  end
    --end
    if bg then
      love.graphics.draw(bg,0,0,0,love.graphics.getWidth()/bg:getWidth(),love.graphics.getHeight()/bg:getHeight())
    end
    love.graphics.scale(scale,scale)
    love.graphics.push()
    love.graphics.translate(0,offset)
    if core.font then
      love.graphics.setFont(core.font)
    end
    for n,s in ipairs(buttons) do
      if n==active then
        if not btncolora then
          love.graphics.setColor(128,128,128,255)
        else
          love.graphics.setColor(unpack(btncolora))
        end
      else
        if not btncoloria then
          love.graphics.setColor(64,64,64,255)
        else
          love.graphics.setColor(unpack(btncoloria))
        end
        if btncolorina and s.ina then
          love.graphics.setColor(unpack(btncolorina))
        end
      end
      if rectbg then
        love.graphics.setRectangleTexture(rectbg)
      end
      if not s.ina then
        love.graphics.rectangle("fill",10,n*200-60,love.graphics.getWidth()/scale-20,200)
      end
      if rectbg then
        love.graphics.setRectangleTexture()
      end
      love.graphics.setColor(255,255,255,255)
      if btntitlecolor then
        love.graphics.setColor(unpack(btntitlecolor))
      else
        love.graphics.setColor(255,255,255,255)end
      love.graphics.printf(buttons[n].title,20,n*200-60,love.graphics.getWidth()/scale/7-20,"center",0,7)
    end
    love.graphics.setBackgroundColor(unpack(bgcolor or nocolor))
    if titlecolor then
      love.graphics.setColor(unpack(titlecolor))
    else
      love.graphics.setColor(255,255,255,255)
    end
    love.graphics.pop()
    love.graphics.printf(title,10,10,love.graphics.getWidth(),"left",0,6)
  end
  while true do
    core.step()
    if func then
      func(buttons,active)
    end
    if End then
      return buttons[active].action,active,offset
    end
  end
end
function core.stopserver(...)
  if core.serverthread and core.serverthread:isRunning() then
    core.serverstarted=false
    local channel=love.thread.getChannel("servernet")
    channel:push("end")
    return true
  end
end
function core.startserver(botcount,map)
  if (not core.serverthread) or (core.serverthread and  not core.serverthread:isRunning()) then
    local thread=love.thread.newThread("server.lua")
    local channel=love.thread.getChannel("servernet")
    thread:start()
    core.serverthread = thread
    core.serverchannel=channel
    channel:clear()
    channel:push(map)
    channel:push(botcount)
    return true
  end
end
